/*
 * xmon - a simple monitor for use
 * on the PowerMac.
 *
 * Copyright (C) 1996 Paul Mackerras.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
#include <setjmp.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef __MWERKS__
#include <MachineExceptions.h>
#include <OSUtils.h>
#include <SIOUX.h>
#else
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#endif

unsigned adrs;
int size = 1;
int atflag = 0;
int mnoread = 0;
unsigned ndump = 64;
unsigned nidump = 16;
unsigned nfdump = 12;
unsigned stride = 1;
int termch;

#ifdef __MWERKS__
jmp_buf blah;
jmp_buf bus_error_jmp;
int catch_bus_error;

typedef unsigned char	uchar;
typedef unsigned short	ushort;
typedef unsigned long	ulong;

OSStatus badboy(ExceptionInformationPowerPC *info);
void cmds(ExceptionInformationPowerPC *);

#else
caddr_t monp;
unsigned monsize;
unsigned monbase;

void cmds(void *);
#endif

int mread(unsigned, void *, int);
int mwrite(unsigned, void *, int);
void byterev(unsigned char *, int);
void atcmd(void);
void slashcmd(void);
int readloc(unsigned, void *, int);
int writeloc(unsigned, void *, int);
void memex(void);
int bsesc(void);
void dump(void);
void prdump(unsigned, int, int, int);
#ifdef __MWERKS__
void prndump(unsigned, int);
int nvreadb(unsigned);
void nvwriteb(unsigned, int);
void nvload(void);
void nvstore(void);
void vcmds(void);
#endif
int ppc_inst_dump(unsigned, int);
int m68k_inst_dump(unsigned, int);
void print_address(unsigned);
void flt_dump(unsigned, unsigned);
void dbl_dump(unsigned, unsigned);
#ifdef __MWERKS__
void excprint(ExceptionInformationPowerPC *);
void prregs(ExceptionInformationPowerPC *);
int getmsr(void);
int getpc(void);
int getsp(void);
int gettoc(void);
#else
void excprint(void *);
void prregs(void *);
#endif
void memops(int);
void memlocate(void);
void memzcan(void);
void memdiffs(unsigned char *, unsigned char *, unsigned, unsigned);
int skipbl(void);
int scanhex(unsigned *valp);
void scannl(void);
int hexdigit(int);
void getstring(char *, int);
void flush_input(void);
int inchar(void);
void take_input(char *);

int print_insn_big_powerpc(FILE *, unsigned long, unsigned);
int print_m68k_insn(FILE *, unsigned long, unsigned char *);

#define GETWORD(v)	(((v)[0] << 24) + ((v)[1] << 16) + ((v)[2] << 8) + (v)[3])

#ifndef __MWERKS__
int fd;
unsigned int offset;
#endif

main(int ac, char **av)
{
#ifdef __MWERKS__
	SIOUXSettings.asktosaveonclose = 0;
	InstallExceptionHandler(badboy);
	setjmp(blah);
	cmds(NULL);
	exit(0);
#else
	char *fname;
	int c;
	int rdonly = 0;
	int endat = 0;

	while ((c = getopt(ac, av, "ro:e:")) >= 0) {
		switch (c) {
		case 'r':
			rdonly = 1;
			break;
		case 'o':
			endat = 0;
			offset = strtol(optarg, NULL, 16);
			break;
		case 'e':
			endat = 1;
			offset = strtol(optarg, NULL, 16);
			break;
		}
	}
	if (optind != ac - 1) {
		fprintf(stderr, "Usage: %s [-r] [-o offset] [-e endloc] file\n");
		exit(1);
	}

	fname = av[optind];
	fd = open(fname, rdonly? O_RDONLY: O_RDWR);
	if (fd < 0) {
		perror(fname);
		exit(1);
	}

	if (endat) {
		struct stat sb;

		if (fstat(fd, &sb) < 0)
			fprintf(stderr, "Warning: can't stat %s\n", fname);
		else
			offset -= sb.st_size;
	}

	cmds(NULL);
	exit(0);
#endif
}

#ifdef __MWERKS__
struct excname {
	char *name;
	int  is_mem_exc;
} exc_names[] = {
	"unknown", 0,
	"illegal instruction", 0,
	"trap", 0,
	"access", 1,
	"unmapped memory", 1,
	"excluded memory", 1,
	"readonly memory", 1,
	"unresolvable page fault", 1,
	"privilege violation", 0,
	"trace", 0,
	"instruction breakpoint", 0,
	"data breakpoint", 0,
	"integer", 0,
	"floating point", 0,
	"stack overflow", 0,
	"termination", 0,
};
#define N_EXC_NAMES	(sizeof(exc_names) / sizeof(exc_names[0]))

OSStatus
badboy(ExceptionInformationPowerPC *info)
{
	int is_mem = 0;
	char *name = NULL;
	
	if (info->theKind < N_EXC_NAMES) {
		is_mem = exc_names[info->theKind].is_mem_exc;
		name = exc_names[info->theKind].name;
		if (is_mem && catch_bus_error) {
			catch_bus_error = 0;
			longjmp(bus_error_jmp, 1);
		}
	}
	excprint(info);
	cmds(info);
	/*longjmp(blah, 1);*/
	return -1;
}
#endif

char *help_string =
"Commands:                                       Modifiers:\n"
"  < [file]             take commands from file    r = byte reverse\n"
"  > [file]             redirect output to file    b = byte\n"
"  @[r][bhw] areg dreg  set addr/data reg addrs    h = halfword (2 bytes)\n"
"  d[@] adr nb          dump bytes                 w = word (4 bytes)\n"
"  dd[@] adr nd         dump double values         @ = indirect access\n"
"  df[@] adr nf         dump float values\n"
"  dh[@] adr nh         dump halfwords\n"
"  di[@] adr ni         dump ppc instructions\n"
"  dw[@] adr nl         dump words\n"
"  m[@] adr [r][bhw]    examine/change memory\n"
"  md adr1 adr2 nbytes  compare two blocks of memory\n"
"  ml adr end val mask  search for words in [adr,end)\n"
"  mm dst src nbytes    move a block of memory\n"
"  ms dst val nbytes    set a block of memory\n"
"  mz adr skip end      find readable bytes in [adr,end)\n"
"  r                    print registers\n"
#ifdef __MWERKS__
"  s                    print pc, sp, toc\n"
#endif
"  x                    exit xmon\n"
;

char *last_cmd;
FILE *out, *ifile;
char fname[128];

/* Command interpreting routine */
void
cmds(excp)
#ifdef __MWERKS__
ExceptionInformationPowerPC *excp;
#else
void *excp;
#endif
{
	int cmd;

	last_cmd = NULL;
	out = stdout;
	ifile = stdin;
	for(;;){
		if (ifile == stdin)
			printf("mon> ");
		fflush(stdout);
		flush_input();
		termch = 0;
		cmd = skipbl();
		if( cmd == '\n' ) {
			if (last_cmd == NULL)
				continue;
			take_input(last_cmd);
			cmd = inchar();
		}
		last_cmd = NULL;
		switch( cmd ){
		case '>':
			if (out != stdout) {
				fclose(out);
				out = stdout;
			}
			cmd = skipbl();
			if (cmd == '\n')
				break;
			termch = cmd;
			getstring(fname, sizeof(fname));
			if ((out = fopen(fname, "w")) == NULL) {
				perror(fname);
				out = stdout;
			}
			break;
		case '<':
			if (ifile != stdin) {
				fclose(ifile);
				ifile = stdin;
			}
			cmd = skipbl();
			if (cmd == '\n')
				break;
			termch = cmd;
			getstring(fname, sizeof(fname));
			if ((ifile = fopen(fname, "r")) == NULL) {
				perror(fname);
				ifile = stdin;
			}
			break;
		case 'm':
			cmd = inchar();
			switch (cmd) {
			case 'm':
			case 's':
			case 'd':
				memops(cmd);
				break;
			case 'l':
				memlocate();
				break;
			case 'z':
				memzcan();
				break;
			default:
				termch = cmd;
				memex();
			}
			break;
		case '@':
			atcmd();
			break;
		case '/':
			slashcmd();
			break;
		case 'd':
			dump();
			break;
		case 'r':
			if (excp != NULL)
				prregs(excp);	/* print regs */
			break;
		case 'e':
			if (excp == NULL)
				printf("No exception information\n");
			else
				excprint(excp);
			break;
#ifdef __MWERKS__
		case 'b':
			Debugger();
			break;
		case 'v':
			vcmds();
			break;
		case 's':
			printf("pc = %x, sp = %x, toc = %x\n", getpc(), getsp(), gettoc());
			break;
#endif
		case EOF:
			if (ifile != stdin) {
				fclose(ifile);
				ifile = stdin;
				break;
			}
			/* fall through */
		case 'x':
			if (out != stdout)
				fclose(out);
			return;
		case '?':
			printf(help_string);
			break;
		default:
			printf("Unrecognized command: ");
			if( ' ' < cmd && cmd <= '~' )
				putchar(cmd);
			else
				printf("\\x%x", cmd);
			printf(" (type ? for help)\n");
		}
	}
}

#ifdef __MWERKS__

asm int
getpc()
{
	mflr	r3
	blr
}

asm int
getsp()
{
	mr		r3,r1
	blr
}

asm int
gettoc()
{
	mr		r3,r2
	blr
}
#endif

/* Commands */

/*
 * Stuff for reading and writing memory safely
 */
int
mread(unsigned adrs, void *buf, int size)
{
#ifdef __MWERKS__
	volatile int n;
	char *p, *q;

	n = 0;
	if( setjmp(bus_error_jmp) == 0 ){
		catch_bus_error = 1;
		p = (char *) adrs;
		q = (char *) buf;
		switch (size) {
		case 2: *(short *)q = *(short *)p;	break;
		case 4: *(int *)q = *(int *)p;		break;
		default:
			for( ; n < size; ++n )
				*q++ = *p++;
		}
		n = size;
	}
	catch_bus_error = 0;
	return n;
#else
	int n;

	if (adrs < offset)
		return 0;
	lseek(fd, adrs - offset, 0);
	n = read(fd, buf, size);
	return n;
#endif
}

int
mwrite(unsigned adrs, void *buf, int size)
{
#ifdef __MWERKS__
	volatile int n;
	char *p, *q;

	n = 0;
	if( setjmp(bus_error_jmp) == 0 ){
		catch_bus_error = 1;
		p = (char *) adrs;
		q = (char *) buf;
		switch (size) {
		case 2: *(short *)p = *(short *)q;	break;
		case 4: *(int *)p = *(int *)q;		break;
		default:
			for( ; n < size; ++n )
				*p++ = *q++;
		}
		n = size;
	} else {
		printf("*** Error writing address %x\n", adrs + n);
	}
	catch_bus_error = 0;
	return n;
#else
	int n;

	if (adrs < offset)
		return 0;
	lseek(fd, adrs - offset, 0);
	n = write(fd, buf, size);
	if (n < 0)
		fprintf(stderr, "*** Error writing address %x\n", adrs);
	return 0;
#endif
}

#define SWAP(a, b, t)	((t) = (a), (a) = (b), (b) = (t))

void
byterev(unsigned char *val, int size)
{
	int t;
	
	switch (size) {
	case 2:
		SWAP(val[0], val[1], t);
		break;
	case 4:
		SWAP(val[0], val[3], t);
		SWAP(val[1], val[2], t);
		break;
	}
}

unsigned addrreg, addrsize, addrrev;
unsigned datareg;

void
atcmd()
{
	int cmd;
	
	cmd = inchar();
	addrrev = 0;
	if (cmd == 'r') {
		addrrev = 1;
		cmd = inchar();
	}
	addrsize = 1;
	switch (cmd) {
	case 'b':	addrsize = 1;	break;
	case 'h':	addrsize = 2;	break;
	case 'w':	addrsize = 4;	break;
	default:	termch = cmd;
	}
	scanhex(&addrreg);
	datareg = addrreg;
	scanhex(&datareg);
}

unsigned slashsize = 16;

void
slashcmd()
{
	if (!scanhex(&slashsize))
		printf("/ step = %x\n", slashsize);
}

int readloc(unsigned adrs, void *ptr, int nb)
{
	unsigned char ab[4];
	int i;
	
	if (!atflag)
		return mread(adrs, ptr, nb);
	for (i = 0; i < addrsize; ++i)
		ab[i] = adrs >> ((addrsize - i - 1) * 8);
	if (addrrev)
		byterev(ab, addrsize);
	if (mwrite(addrreg, ab, addrsize) != addrsize)
		return 0;
	return mread(datareg, ptr, nb);
}

int writeloc(unsigned adrs, void *ptr, int nb)
{
	unsigned char ab[4];
	int i;
	
	if (!atflag)
		return mwrite(adrs, ptr, nb);
	for (i = 0; i < addrsize; ++i)
		ab[i] = adrs >> ((addrsize - i - 1) * 8);
	if (addrrev)
		byterev(ab, addrsize);
	if (mwrite(addrreg, ab, addrsize) != addrsize)
		return 0;
	return mwrite(datareg, ptr, nb);
}

void
memex()
{
	int cmd, inc, i, sstep;
	unsigned n;
	unsigned char val[4];
	static int brev = 0;
	
	cmd = skipbl();
	atflag = 0;
	if (cmd == '@') {
		atflag = 1;
		last_cmd = "m@\n";
	} else {
		termch = cmd;
		last_cmd = "m\n";
	}
	scanhex(&adrs);
	while ((cmd = skipbl()) != '\n') {
		switch( cmd ){
		case 'b':	size = 1;	break;
		case 'w':	size = 2;	break;
		case 'l':	size = 4;	break;
		case 'r': 	brev = !brev;	break;
		case 'n':	mnoread = 1;	break;
		case '.':	mnoread = 0;	break;
		}
	}
	if( size <= 0 )
		size = 1;
	else if( size > 4 )
		size = 4;
	if( cmd != '\n' )
		scannl();
	for(;;){
		if (!mnoread)
			n = readloc(adrs, val, size);
		printf("%.*x%s ", (atflag? addrsize*2: 8), adrs, (brev? "r": ""));
		if (!mnoread) {
			if (brev)
				byterev(val, size);
			for (i = 0; i < n; ++i)
				printf("%.2x", val[i]);
			for (; i < size; ++i) {
				printf("**");
				val[i] = 0;
			}
		}
		putchar(' ');
		inc = size;
		sstep = 0;
		for(;;){
			if (scanhex(&n)) {
				for (i = 0; i < size; ++i)
					val[i] = n >> ((size - i - 1) * 8);
				if (brev)
					byterev(val, size);
				writeloc(adrs, val, size);
				inc = size;
			}
			cmd = skipbl();
			if (cmd == '\n')
				break;
			inc = 0;
			switch (cmd) {
			case '\'':
				for(;;){
					n = inchar();
					if( n == '\\' )
						n = bsesc();
					else if( n == '\'' )
						break;
					for (i = 0; i < size; ++i)
						val[i] = n >> (i * 8);
					if (!brev)
						byterev(val, size);
					writeloc(adrs, val, size);
					adrs += size;
				}
				adrs -= size;
				inc = size;
				break;
			case ',':
				adrs += size;
				break;
			case '.':
				mnoread = 0;
				break;
			case ';':
				break;
			case 'x':
			case EOF:
				scannl();
				return;
			case 'b':
			case 'v':
				size = 1;
				break;
			case 'h':
				size = 2;
				break;
			case 'w':
				size = 4;
				break;
			case '^':
				adrs -= size;
				break;
			case '/':
				if (sstep > 0)
					adrs -= sstep;
				else
					sstep = 1;
				sstep *= slashsize;
				adrs += sstep;
				break;
			case '\\':
				if (sstep < 0)
					adrs -= sstep;
				else
					sstep = -1;
				sstep *= slashsize;
				adrs += sstep;
				break;
			case 'm':
				scanhex(&adrs);
				break;
			case 'n':
				mnoread = 1;
				break;
			case 'r':
				brev = !brev;
				break;
			case '<':
				n = size;
				scanhex(&n);
				adrs -= n;
				break;
			case '>':
				n = size;
				scanhex(&n);
				adrs += n;
				break;
			case '*':
				adrs = 0;
				for (i = 0; i < size; ++i)
					adrs = (adrs << 8) + val[i];
				break;
			}
		}
		adrs += inc;
	}
}

int
bsesc()
{
	int c;

	c = inchar();
	switch( c ){
	case 'n':	c = '\n';	break;
	case 'r':	c = '\r';	break;
	case 'b':	c = '\b';	break;
	case 't':	c = '\t';	break;
	}
	return c;
}

void
dump()
{
	int c, x, size;
	unsigned n;
	static char last[8];
	char *p = last;

	*p++ = 'd';
	atflag = 0;
	c = inchar();
	if (c == '@') {
		atflag = 1;
		c = inchar();
	}
	if (isxdigit(c) && c != 'f' && c != 'd' || c == '\n') {
		termch = c;
	} else {
		*p++ = c;
		x = inchar();
		if (x == '@')
			atflag = 1;
		else
			termch = x;
	}
	if (atflag)
		*p++ = '@';
	*p++ = '\n';
	*p++ = 0;
	last_cmd = last;
	scanhex(&adrs);
	if( termch != '\n')
		termch = 0;
	if( c == 'i' ){
		scanhex(&nidump);
		if( nidump == 0 )
			nidump = 16;
		adrs += ppc_inst_dump(adrs, nidump);
	} else if( c == 'k' ){
		scanhex(&nidump);
		if( nidump == 0 )
			nidump = 16;
		adrs += m68k_inst_dump(adrs, nidump);
	} else if( c == 'f' ){
		scanhex(&nfdump);
		if( nfdump == 0 )
			nfdump = 12;
		flt_dump(adrs, nfdump);
		adrs += nfdump * sizeof(float);
	} else if( c == 'd' ){
		scanhex(&nfdump);
		if( nfdump == 0 )
			nfdump = 12;
		dbl_dump(adrs, nfdump);
		adrs += nfdump * sizeof(double);
#ifdef __MWERKS__
	} else if (c == 'n') {
		scanhex(&ndump);
		prndump(adrs, ndump);
		adrs += ndump;
#endif
	} else {
		size = (c == 'w')? 4: (c == 'h')? 2: 1;
		if (scanhex(&n))
			ndump = n * size;
		if( ndump == 0 )
			ndump = 64;
		stride = size;
		scanhex(&stride);
		prdump(adrs, ndump, stride, size);
		adrs += ndump;
	}
}

#ifdef __MWERKS__
#define NVADRS	((volatile char *)0xF301D000)
#define NVDATA	((volatile char *)0xF301F000)

asm void eieio()
{
	eieio
	blr
}

int
nvreadb(unsigned adr)
{
	*NVADRS = adr >> 5;
	eieio();
	return NVDATA[(adr & 0x1f) << 4];
}

void
nvwriteb(unsigned adr, int v)
{
	*NVADRS = adr >> 5;
	eieio();
	NVDATA[(adr & 0x1f) << 4] = v;
}

#define NVSTART	0x1800
#define NVSIZE	0x800

struct nvram {
	unsigned short magic;		/* 0x1275 */
	unsigned char  char2;
	unsigned char  char3;
	unsigned short cksum;
	unsigned short end_vals;
	unsigned short start_strs;
	unsigned short word5;
	unsigned long  bits;
	unsigned long  vals[1];
};

union {
	struct nvram nv;
	char c[NVSIZE];
	unsigned short s[NVSIZE/2];
} nvbuf;

int
nvcsum()
{
	int i;
	unsigned c;
	
	c = 0;
	for (i = 0; i < NVSIZE/2; ++i)
		c += nvbuf.s[i];
	c = (c & 0xffff) + (c >> 16);
	c += (c >> 16);
	return c & 0xffff;
}

void
nvload()
{
	int i;
	int s;
	
	for (i = 0; i < NVSIZE; ++i)
		nvbuf.c[i] = nvreadb(i + NVSTART);
	s = nvcsum();
	if (s != 0xffff)
		printf("checksum = %x\n", nvcsum());
}

void
nvstore()
{
	int i;
	
	for (i = 0; i < NVSIZE; ++i)
		nvwriteb(i + NVSTART, nvbuf.c[i]);
}

enum nvtype {
	boolean,
	word,
	string
};

struct nvvar {
	char *name;
	enum nvtype type;
} nvvars[] = {
	"little-endian?", boolean,
	"real-mode?", boolean,
	"auto-boot?", boolean,
	"diag-switch?", boolean,
	"fcode-debug?", boolean,
	"oem-banner?", boolean,
	"oem-logo?", boolean,
	"use-nvramrc?", boolean,
	"real-base", word,
	"real-size", word,
	"virt-base", word,
	"virt-size", word,
	"load-base", word,
	"pci-probe-list", word,
	"screen-#columns", word,
	"screen-#rows", word,
	"selftest-#megs", word,
	"boot-device", string,
	"boot-file", string,
	"diag-device", string,
	"diag-file", string,
	"input-device", string,
	"output-device", string,
	"oem-banner", string,
	"oem-logo", string,
	"nvramrc", string,
	"boot-command", string,
};

#define N_NVVARS	(sizeof(nvvars) / sizeof(nvvars[0]))

union nvval {
	unsigned int word_val;
	char *str_val;
} nvvals[32];

char nvstrbuf[NVSIZE];
int nvstr_used;

void nvunpack()
{
	int i;
	unsigned long bmask;
	int vi, off, len;

	nvstr_used = 0;	
	bmask = 0x80000000;
	vi = 0;
	for (i = 0; i < N_NVVARS; ++i) {
		switch (nvvars[i].type) {
		case boolean:
			nvvals[i].word_val = (nvbuf.nv.bits & bmask)? 1: 0;
			bmask >>= 1;
			break;
		case word:
			nvvals[i].word_val = nvbuf.nv.vals[vi++];
			break;
		case string:
			off = nvbuf.nv.vals[vi] >> 16;
			len = nvbuf.nv.vals[vi++] & 0xffff;
			nvvals[i].str_val = nvstrbuf + nvstr_used;
			memcpy(nvvals[i].str_val, nvbuf.c + off - NVSTART, len);
			nvvals[i].str_val[len] = 0;
			nvstr_used += len + 1;
			break;
		}
	}
}

void
nvpack()
{
	int i, off, len, vi;
	unsigned long bmask;
	
	bmask = 0x80000000;
	vi = 0;
	off = NVSIZE;
	nvbuf.nv.bits = 0;
	for (i = 0; i < N_NVVARS; ++i) {
		switch (nvvars[i].type) {
		case boolean:
			if (nvvals[i].word_val)
				nvbuf.nv.bits |= bmask;
			bmask >>= 1;
			break;
		case word:
			nvbuf.nv.vals[vi++] = nvvals[i].word_val;
			break;
		case string:
			len = strlen(nvvals[i].str_val);
			off -= len;
			memcpy(nvbuf.c + off, nvvals[i].str_val, len);
			nvbuf.nv.vals[vi++] = ((off + NVSTART) << 16) + len;
			break;
		}
	}
	nvbuf.nv.magic = 0x1275;
	nvbuf.nv.cksum = 0;
	nvbuf.nv.end_vals = NVSTART + (unsigned) &nvbuf.nv.vals[vi] - (unsigned) &nvbuf;
	nvbuf.nv.start_strs = off + NVSTART;
	memset(&nvbuf.c[nvbuf.nv.end_vals - NVSTART], 0, nvbuf.nv.start_strs - nvbuf.nv.end_vals);
	nvbuf.nv.cksum = ~nvcsum();
}

unsigned char nvsave[0x2000];

void
vcmds()
{
	int i, c, cmd, x;
	char *str, *p;
	char vname[33];
	
	nvload();
	nvunpack();
	cmd = skipbl();
	switch (cmd) {
	case '\n':
		for (i = 0; i < N_NVVARS; ++i) {
			printf("%-16s", nvvars[i].name);
			switch (nvvars[i].type) {
			case boolean:
				printf("%s", nvvals[i].word_val? "true": "false");
				break;
			case word:
				printf("0x%x", nvvals[i].word_val);
				break;
			case string:
				for (p = nvvals[i].str_val; *p != 0; ++p)
					if (*p != '\r')
						putchar(*p);
					else
						printf("\n                ");
				break;
			}
			printf("\n");
		}
		break;
	case 's':
		getstring(vname, sizeof(vname));
		for (i = 0; i < N_NVVARS; ++i) {
			if (strcmp(vname, nvvars[i].name) == 0)
				break;
		}
		if (i >= N_NVVARS) {
			printf("no variable called '%s'\n", vname);
			scannl();
			break;
		}
		c = skipbl();
		p = str = nvstrbuf + nvstr_used;
		while (c != '\n') {
			*p++ = c;
			c = inchar();
		}
		*p++ = 0;
		switch (nvvars[i].type) {
		case boolean:
			if (strcmp(str, "true") == 0)
				nvvals[i].word_val = 1;
			else if (strcmp(str, "false") == 0)
				nvvals[i].word_val = 0;
			else
				printf("bad boolean value '%s' for %s\n", str, vname);
			break;
		case word:
			nvvals[i].word_val = strtoul(str, 0, 16);
			break;
		case string:
			nvvals[i].str_val = str;
			nvstr_used += p - str;
			break;
		}
		nvpack();
		nvstore();
		break;
	case 'c':
		for (i = 0; i < sizeof(nvsave); ++i)
			nvsave[i] = nvreadb(i);
		break;
	case 'd':
		for (i = 0; i < sizeof(nvsave); ++i) {
			x = nvreadb(i) & 0xff;
			if (x != nvsave[i])
				printf("%4x was %2x is %2x\n", i, nvsave[i], x);
		}
		break;
	default:
		printf("unknown command v%c\n", cmd);
		scannl();
	}
}

void
prndump(unsigned adrs, int ndump)
{
	register int n, m, c, r;
	unsigned char temp[16];

	for( n = ndump; n > 0; ){
		printf("%.8x ", adrs);
		r = n < 16? n: 16;
		for( m = 0; m < r; ++m ){
			temp[m] = nvreadb(adrs + m);
			putchar((m & 3) == 0 & m > 0? '.': ' ');
			printf("%.2x", temp[m]);
		}
		for(; m < 16; ++m )
			printf("   ");
		printf("  |");
		for( m = 0; m < r; ++m ){
			c = temp[m];
			putchar(' ' <= c && c <= '~'? c: '.');
		}
		adrs += r;
		n -= r;
		for(; m < 16; ++m )
			putchar(' ');
		printf("|\n");
	}
}
#endif

void
prdump(unsigned adrs, int ndump, int stride, int isize)
{
	register int n, m, c, r, nr, nok;
	unsigned char temp[16], ok[16];

	for( n = ndump; n > 0; ){
		fprintf(out, "%.8x ", adrs);
		r = n < 16? n: 16;
		memset(ok, 0, sizeof(ok));
		for (nr = 0; nr < r; nr += isize) {
			nok = readloc(adrs, &temp[nr], isize);
			memset(ok+nr, 1, nok);
			adrs += stride;
		}
		for( m = 0; m < 16; ++m ){
			if (m > 0 && m % isize == 0) {
				if (isize == 4)
					fprintf(out, "  ");
				else if (isize == 2)
					fprintf(out, (m & 3) == 0? "  ": " ");
				else
					fprintf(out, (m & 3) == 0? ".": " ");
			}
			if (m < r)
				fprintf(out, (ok[m]? "%.2x": "**"), temp[m]);
			else
				fprintf(out, "  ");
		}
		fprintf(out, "  |");
		for( m = 0; m < r; ++m ){
			if (m < r && ok[m]) {
				c = temp[m];
				putc(' ' <= c && c <= '~'? c: '.', out);
			} else
				putc(' ', out);
		}
		n -= r;
		for (; m < 16; ++m)
			putc(' ', out);
		fprintf(out, "|\n");
	}
}

int
ppc_inst_dump(unsigned adr, int count)
{
	int nr, dotted;
	unsigned first_adr;
	unsigned long inst, last_inst;
	unsigned char val[4];

	dotted = 0;
	for (first_adr = adr; count > 0; --count, adr += 4){
		nr = readloc(adr, val, 4);
		if( nr == 0 ){
			fprintf(out, "%.8x  ********\n", adr);
			break;
		}
		inst = GETWORD(val);
		if (adr > first_adr && inst == last_inst) {
			if (!dotted) {
				fprintf(out, " ...\n");
				dotted = 1;
			}
			continue;
		}
		dotted = 0;
		last_inst = inst;
		fprintf(out, "%.8x  ", adr);
		fprintf(out, "%.8x\t", inst);
		print_insn_big_powerpc(out, inst, adr);	/* always returns 4 */
		fprintf(out, "\n");
	}
	return adr - first_adr;
}

int
m68k_inst_dump(unsigned adr, int count)
{
	int i, j, size, nr;
	unsigned first_adr;
	unsigned char inst[22];

	first_adr = adr;
	for( ; count > 0; --count ){
		fprintf(out, "%.8x  ", adr);
		memset(inst, 0, sizeof(inst));
		nr = readloc(adr, inst, sizeof(inst));
		if( nr == 0 ){
			fprintf(out, "****\n");
			break;
		}
		size = print_m68k_insn(NULL, (unsigned long)adr, inst);
		if( size == 0 ){
			fprintf(out, "%.2x%.2x\n", inst[0], inst[1]);
			adr += 2;
		} else {
			for( i = 0; i < 6 && i < size; i += 2 ){
				if( i + 1 < nr )
					fprintf(out, "%.2x%.2x ", inst[i], inst[i+1]);
				else
					fprintf(out, "**** ");
			}
			for( j = i; j < 6; j += 2 )
				fprintf(out, "     ");
			fprintf(out, "\t");
			size = print_m68k_insn(out, (unsigned long)adr, inst);
			if( i < size ){
				fprintf(out, "\n         ");
				for( ; i < size; i += 2 ){
				    fprintf(out, " ");
				    if( i + 1 < nr )
						fprintf(out, "%.2x%.2x", inst[i], inst[i+1]);
				    else
						fprintf(out, "****");
				}
			}
			adr += size;
			fprintf(out, "\n");
			if( nr < size )
				break;
		}
	}
	return adr - first_adr;
}

void
print_address(addr)
unsigned addr;
{
	fprintf(out, "0x%x", addr);
}

void
flt_dump(unsigned adrs, unsigned count)
{
	unsigned n;
	float f;

	for( n = 0; n < count; ++n ){
		if( n % 4 == 0 )
			fprintf(out, "%.8x  ", adrs);
		if( readloc(adrs, &f, sizeof(f)) < sizeof(f) ){
			fprintf(out, "******\n");
			break;
		}
		fprintf(out, "%16.5f", f);
		if( n % 4 == 3 || n + 1 == count )
			fprintf(out, "\n");
		adrs += sizeof(f);
	}
}

void
dbl_dump(unsigned adrs, unsigned count)
{
	unsigned n;
	double f;

	for( n = 0; n < count; ++n ){
		if( n % 3 == 0 )
			fprintf(out, "%.8x  ", adrs);
		if( readloc(adrs, &f, sizeof(f)) < sizeof(f) ){
			fprintf(out, "******\n");
			break;
		}
		fprintf(out, "%22.14f", f);
		if( n % 3 == 2 || n + 1 == count )
			fprintf(out, "\n");
		adrs += sizeof(f);
	}
}

#ifdef __MWERKS__
/* Describe the exception */
void
excprint(info)
ExceptionInformationPowerPC *info;
{
	int is_mem = 0;
	char *name = NULL;
	
	if (info->theKind < N_EXC_NAMES) {
		is_mem = exc_names[info->theKind].is_mem_exc;
		name = exc_names[info->theKind].name;
	}
	if (name != NULL)
		printf("%s exception\n", name);
	else
		printf("?exception (kind = %d)\n", info->theKind);
	if (is_mem) {
		printf("  memory area=%x address=%x error=%x ref=%x\n",
			   info->info.memoryInfo->theArea,
			   info->info.memoryInfo->theAddress,
			   info->info.memoryInfo->theError,
			   info->info.memoryInfo->theReference);
	}
}

/* Print registers */
void
prregs(info)
ExceptionInformationPowerPC *info;
{
	int n;
	UnsignedWide *rp;
	
	rp = &info->registerImage->R0;
	for (n = 0; n < 32; ++n, ++rp) {
		printf("R%.2d = %.8x%s", n, rp->lo, (n & 3) == 3? "\n": "   ");
	}
	printf("ctr = %.8x    lr = %.8x    pc = %.8x    cr = %.8x\n",
		   info->machineState->CTR.lo, info->machineState->LR.lo,
		   info->machineState->PC.lo, info->machineState->CR);
	printf("xer = %.8x   msr = %.8x\n",
		   info->machineState->XER, info->machineState->MSR);
	printf("fpscr = %x\n", info->FPUImage->FPSCR);
}

asm int
getmsr()
{
	mfmsr	r3
	blr
}

#else
/* Describe the exception */
void
excprint(info)
void *info;
{
}

/* Print registers */
void
prregs(info)
void *info;
{
}
#endif

/*
 * Memory operations - move, set, print differences
 */
unsigned mdest;		/* destination address */
unsigned msrc;		/* source address */
unsigned mval;		/* byte value to set memory to */
unsigned mcount;	/* # bytes to affect */
unsigned mdiffs;	/* max # differences to print */

void
memops(int cmd)
{
	scanhex(&mdest);
	if( termch != '\n' )
		termch = 0;
	scanhex(cmd == 's'? &mval: &msrc);
	if( termch != '\n' )
		termch = 0;
	scanhex(&mcount);
	switch( cmd ){
	case 'm':
		memmove((void *)mdest, (void *)msrc, mcount);
		break;
	case 's':
		memset((void *)mdest, mval, mcount);
		break;
	case 'd':
		if( termch != '\n' )
			termch = 0;
		scanhex(&mdiffs);
		memdiffs((unsigned char *)mdest, (unsigned char *)msrc, mcount, mdiffs);
		break;
	}
}

void
memdiffs(unsigned char *p1, unsigned char *p2, unsigned nb, unsigned maxpr)
{
	unsigned n, prt;

	prt = 0;
	for( n = nb; n > 0; --n )
		if( *p1++ != *p2++ )
			if( ++prt <= maxpr )
				printf("%.8x %.2x # %.8x %.2x\n", (unsigned)p1 - 1,
					p1[-1], (unsigned)p2 - 1, p2[-1]);
	if( prt > maxpr )
		printf("Total of %d differences\n", prt);
}

unsigned mend;
unsigned mask;

void
memlocate()
{
	unsigned a, n;
	unsigned char val[4];

	last_cmd = "ml";
	scanhex(&mdest);
	if (termch != '\n') {
		termch = 0;
		scanhex(&mend);
		if (termch != '\n') {
			termch = 0;
			scanhex(&mval);
			mask = ~0;
			if (termch != '\n') termch = 0;
			scanhex(&mask);
		}
	}
	n = 0;
	for (a = mdest; a < mend; a += 4) {
		if (mread(a, val, 4) == 4
			&& ((GETWORD(val) ^ mval) & mask) == 0) {
			printf("%.8x:  %.8x\n", a, GETWORD(val));
			if (++n >= 10)
				break;
		}
	}
}

unsigned mskip = 0x1000;
unsigned mlim = 0xffffffff;

void
memzcan()
{
	unsigned char v;
	unsigned a;
	int ok, ook;

	scanhex(&mdest);
	if (termch != '\n') termch = 0;
	scanhex(&mskip);
	if (termch != '\n') termch = 0;
	scanhex(&mlim);
	ook = 0;
	for (a = mdest; a < mlim; a += mskip) {
		ok = mread(a, &v, 1);
		if (ok && !ook) {
			printf("%.8x .. ", a);
			fflush(stdout);
		} else if (!ok && ook)
			printf("%.8x\n", a - mskip);
		ook = ok;
		if (a + mskip < a)
			break;
	}
	if (ook)
		printf("%.8x\n", a - mskip);
}

/* Input scanning routines */
int
skipbl()
{
	int c;

	if( termch != 0 ){
		c = termch;
		termch = 0;
	} else
		c = inchar();
	while( c == ' ' || c == '\t' )
		c = inchar();
	return c;
}

int
scanhex(vp)
unsigned *vp;
{
	int c, d;
	unsigned v;

	c = skipbl();
	d = hexdigit(c);
	if( d == EOF ){
		termch = c;
		return 0;
	}
	v = 0;
	do {
		v = (v << 4) + d;
		c = inchar();
		d = hexdigit(c);
	} while( d != EOF );
	termch = c;
	*vp = v;
	return 1;
}

void
scannl()
{
	int c;

	c = termch;
	termch = 0;
	while( c != '\n' )
		c = inchar();
}

int
hexdigit(c)
{
	if( '0' <= c && c <= '9' )
		return c - '0';
	if( 'A' <= c && c <= 'F' )
		return c - ('A' - 10);
	if( 'a' <= c && c <= 'f' )
		return c - ('a' - 10);
	return EOF;
}

void
getstring(char *s, int size)
{
	int c;

	c = skipbl();
	do {
		if( size > 1 ){
			*s++ = c;
			--size;
		}
		c = inchar();
	} while( c != ' ' && c != '\t' && c != '\n' );
	termch = c;
	*s = 0;
}

char line[256];
char *lineptr;

void
flush_input()
{
	lineptr = NULL;
}

int
inchar()
{
	if (lineptr == NULL || *lineptr == 0) {
		if (fgets(line, sizeof(line), ifile) == NULL) {
			lineptr = NULL;
			return EOF;
		}
		if (ifile != stdin || out != stdout)
			fputs(line, out);
		lineptr = line;
	}
	return *lineptr++;
}

void
take_input(str)
char *str;
{
	lineptr = str;
}
